
#include <Ogre.h>

#include "PlayState.h"
#include "GuiManager.h"
#include "OgreConsole.h"
#include "ClientDataManager.h"
#include "SpringCam.h"

using namespace Ogre;

PlayState * PlayState::mPlayState;

// ============================================================================
void PlayState::update( unsigned long lTimeElapsedInMilis )
{
	// Process input
	ClientDataManager::getSingleton().processInput(lTimeElapsedInMilis);

	// Update logic and networking
	ClientDataManager::getSingleton().update(lTimeElapsedInMilis);

	// Update particles
	if(ClientDataManager::getSingleton().getLocalPlayer().isInitialized())
	{
		ClientDataManager::getSingleton().getLocalPlayer().particleUpdate();

		// Update GUI
		stiffnessText->setCaption(Ogre::StringConverter::toString(
			stiffnessScroll->getScrollPosition()+1));
		dampeningText->setCaption(Ogre::StringConverter::toString(
			dampeningScroll->getScrollPosition()+1));
		massText->setCaption(Ogre::StringConverter::toString(
			massScroll->getScrollPosition()+1));

		// Set the camera to look from our local client
		float timeSinceInSeconds = (float)lTimeElapsedInMilis / 1000;

		Vector3 playerPos = ClientDataManager::getSingleton().getLocalPlayer().getPosition();

		Vector3 facingDir = ClientDataManager::getSingleton().getLocalPlayer().getFacingDirection();
		facingDir.normalise();

		Vector3 upDir = ClientDataManager::getSingleton().getLocalPlayer().getUpDirection();
		upDir.normalise();

		Vector3 camDesiredPos = 300 * facingDir 
			// shift up, above the player
			+ 50 * upDir
			+ playerPos;

		Vector3 camCurrentShiftFromDesired = mCamera->getPosition() - camDesiredPos;
		// Don't stretch TOO much 
		if(camCurrentShiftFromDesired.length() > 200.0f)
		{
			Vector3 shiftDir = camCurrentShiftFromDesired.normalisedCopy();
			mCamera->setPosition(camDesiredPos + shiftDir * 200.0f);
		}
		Vector3 camLookingTarget = -50 * facingDir + playerPos;



		float stiffness, dampening, mass;
		stiffness = stiffnessScroll->getScrollPosition()+1;
		dampening = dampeningScroll->getScrollPosition()+1;
		mass = massScroll->getScrollPosition()+1;

		mCamera->setPosition(SpringCam::getInstance()->updateSpringCam(mCamera->getPosition(),camDesiredPos,timeSinceInSeconds,stiffness,dampening,mass));

		
		// Wanted the camera to be looking at a certain point, but also to retain it the initial
		// player Roll angle. After lots of testing and tweaking with camera->lookAt() + custom Quaternions,
		// finally switched to pure quaternion equations and eventually succeeded 

		// Get the initial orientation
		Quaternion newOrientation = ClientDataManager::getSingleton().getLocalPlayer().getOrientation();
		// Align the camera with player spaceship orientation (rotation)
		mCamera->setOrientation(newOrientation);

		// This is the trick: get the rotation needed from current point to the lookingTarget
		Quaternion flatCameraRotationQuat = mCamera->getDirection().getRotationTo(
			Vector3(camLookingTarget - mCamera->getPosition()).normalisedCopy() );
		// Apply it
		mCamera->rotate(flatCameraRotationQuat); 
		

		// This was harder than in SolarBound, because here we can rotate in full 3D, including
		// custom pitch, roll and yaw.


		coordsText->setCaption(String("Coords:")
 +String(" X: ")+StringConverter::toString(ClientDataManager::getSingleton().getLocalPlayer().getPosition().x)
 +String(" Y: ")+StringConverter::toString(ClientDataManager::getSingleton().getLocalPlayer().getPosition().y)
 +String(" Z: ")+StringConverter::toString(ClientDataManager::getSingleton().getLocalPlayer().getPosition().z)
 );

		livesText->setCaption("Lives: " + StringConverter::toString( ClientDataManager::getSingleton().getLocalPlayer().getLives() ));
		scoreText->setCaption("Score: " + StringConverter::toString( ClientDataManager::getSingleton().getLocalPlayer().getScore() ));
		if (ClientDataManager::getSingleton().getLocalPlayer().getZVelocity() >= 95) speedText->setCaption("Speed: 100"); // too stop it stuttering when value is near 100
		else speedText->setCaption("Speed: " + StringConverter::toString( ClientDataManager::getSingleton().getLocalPlayer().getZVelocity() ));
	} // if(ClientDataManager::getSingleton().getLocalPlayer().isInitialized())

}
// ============================================================================
void PlayState::setupScene()
{
	mSceneMgr = mRoot->createSceneManager(ST_EXTERIOR_CLOSE, "PlayState SceneManager");
	mCamera = mSceneMgr->createCamera("Camera");
	mViewport = mRoot->getAutoCreatedWindow()->addViewport(mCamera);

	// Scene:
	mSceneMgr->setAmbientLight(ColourValue(1.0f, 1.0f, 1.0f));
	mSceneMgr->setSkyBox(true,"Examples/SpaceSkyBox");


}
// ============================================================================
void PlayState::enter( void )
{
	mIsActive = true;

	srand((unsigned)time(0));
	mRoot = Root::getSingletonPtr();

	// -----------------------------------
	this->setupScene();  // important call
	// -----------------------------------

	mSceneMgr = mRoot->getSceneManager( "PlayState SceneManager" );

	// Its easier to write 'gui' than 'GuiManager::getSingleton().getGuiPtr()'
	MyGUI::Gui * gui = GuiManager::getSingleton().getGuiPtr();

	// Now this line took lots and lots of time to debug, combined with destroySceneManager
	gui->setSceneManager(mSceneMgr);
	gui->hidePointer();

	// OgreConsole
	//OgreConsole::getSingleton().init(mRoot);
	OgreConsole::getSingleton().setVisible(false);

	//mRoot->destroySceneManager(mRoot->getSceneManager("IntroState SceneManager"));
	
	// It seems that switching a SceneManager actually needs a WORKING COPY of the previous
	// manager. This idea is lost to me. -________-

	widgetPtr = MyGUI::LayoutManager::getInstance().load("ClientPlayState.layout");
	gui->findWidget<MyGUI::Widget>("Background")->setSize(gui->getViewSize());

	// hiding the background (and thus spring camera settings)
	gui->findWidget<MyGUI::Widget>("Background")->setVisible(false);

	massScroll = gui->findWidget<MyGUI::HScroll>("MassScroll");
	massText = gui->findWidget<MyGUI::StaticText>("MassText");

	dampeningScroll = gui->findWidget<MyGUI::HScroll>("DampeningScroll");
	dampeningText = gui->findWidget<MyGUI::StaticText>("DampeningText");

	stiffnessScroll = gui->findWidget<MyGUI::HScroll>("StiffnessScroll");
	stiffnessText = gui->findWidget<MyGUI::StaticText>("StiffnessText");

	// These controls are separate from the Background widget:
	livesText = gui->createWidget<MyGUI::StaticText>("StaticText", 50, 30, 130, 30, MyGUI::ALIGN_DEFAULT, "Main");
	unsigned int height = static_cast<MyGUI::FontPtr>(
		MyGUI::FontManager::getInstance().getByName("DejaVuSans") )->getDefaultHeight();
	livesText->setFontName("DejaVuSans");
	livesText->setFontHeight(height);
	livesText->setTextAlign( MyGUI::Align::Center );
	livesText->setTextColour(MyGUI::Colour(1.0f,1.0f,1.0f));
	livesText->setCaption("Lives: " + StringConverter::toString( ClientDataManager::getSingleton().getLocalPlayer().getLives() ));

	scoreText = gui->createWidget<MyGUI::StaticText>("StaticText", 50, 60, 130, 30, MyGUI::ALIGN_DEFAULT, "Main");
	scoreText->setFontName("DejaVuSans");
	scoreText->setFontHeight(height);
	scoreText->setTextAlign( MyGUI::Align::Center );
	scoreText->setTextColour(MyGUI::Colour(1.0f,1.0f,1.0f));
	scoreText->setCaption("Score: " + StringConverter::toString( ClientDataManager::getSingleton().getLocalPlayer().getScore() ));
////////////////////////////////////////////////////
	speedText = gui->createWidget<MyGUI::StaticText>("StaticText", 50, 90, 130, 30, MyGUI::ALIGN_DEFAULT, "Main");
	speedText->setFontName("DejaVuSans");
	speedText->setFontHeight(height);
	speedText->setTextAlign( MyGUI::Align::Center );
	speedText->setTextColour(MyGUI::Colour(0.0f,1.0f,0.0f));
	speedText->setCaption("Speed: " + StringConverter::toString( ClientDataManager::getSingleton().getLocalPlayer().getZVelocity() ));
/////////////////////////////////////////////////////

	coordsText = gui->createWidget<MyGUI::StaticText>("StaticText", 50, 500, 400, 66, MyGUI::ALIGN_DEFAULT, "Main");
	coordsText->setFontName("DejaVuSans");
	coordsText->setFontHeight(height);
	coordsText->setTextAlign( MyGUI::Align::Center );
	coordsText->setTextColour(MyGUI::Colour(1.0f,1.0f,1.0f));
	coordsText->setCaption("Coords: ");
	coordsText->setVisible(false);

	deadReckoningText = gui->createWidget<MyGUI::StaticText>("StaticText", 400, 20, 300, 20, MyGUI::ALIGN_DEFAULT, "Main");
	deadReckoningText->setFontName("DejaVuSans");
	deadReckoningText->setFontHeight(height);
	deadReckoningText->setTextAlign( MyGUI::Align::Left );
	deadReckoningText->setTextColour(MyGUI::Colour(1.0f,1.0f,1.0f));

	string deadActive = ClientDataManager::getSingleton().isDeadReckoningActive?"activated":"deactivated";
	deadReckoningText->setCaption("Dead Reckoning ('R'): " + deadActive);

	
}
// ============================================================================
void PlayState::exit( void )
{
	GuiManager::getSingleton().getGuiPtr()->destroyWidget(livesText);
	GuiManager::getSingleton().getGuiPtr()->destroyWidget(scoreText);
	GuiManager::getSingleton().getGuiPtr()->destroyWidget(coordsText);
	GuiManager::getSingleton().getGuiPtr()->destroyWidget(deadReckoningText);

	MyGUI::LayoutManager::getInstance().unloadLayout(widgetPtr);

	mRoot->getAutoCreatedWindow()->removeViewport(0);

	SceneManager* killMe = mRoot->getSceneManager("PlayState SceneManager");
	mRoot->destroySceneManager(killMe);

	mIsActive = false;
}
// ============================================================================
void PlayState::pause( void )
{
}
// ============================================================================
void PlayState::resume( void )
{
}

// ============================================================================
void PlayState::keyPressed( const OIS::KeyEvent &e )
{
	MyGUI::Gui * gui = GuiManager::getSingleton().getGuiPtr();
	// Switch spring camera GUI visibility by pressing '/'
	if( e.key == OIS::KC_SLASH)
	{
		bool visible = gui->findWidget<MyGUI::Widget>("Background")->isVisible();
		gui->findWidget<MyGUI::Widget>("Background")->setVisible(!visible);
		if(visible)
		{
			gui->hidePointer();
		}
		else
		{
			gui->showPointer();
		}
	}

	// show/hide console/debug windows/info:
	if( e.key == OIS::KC_GRAVE || e.key == OIS::KC_KANJI) // keys: ~ or `
	{
		bool visible = OgreConsole::getSingleton().isVisible();
		OgreConsole::getSingleton().setVisible( !visible );
		coordsText->setVisible( !visible );
	}

	if(e.key == OIS::KC_R)
	{
		// switch dead reckoning
		ClientDataManager::getSingleton().isDeadReckoningActive = 
			!ClientDataManager::getSingleton().isDeadReckoningActive;

		string deadActive = ClientDataManager::getSingleton().isDeadReckoningActive?"activated":"deactivated";
		deadReckoningText->setCaption("Dead Reckoning ('R'): " + deadActive);
	}
}
// ============================================================================
void PlayState::keyReleased( const OIS::KeyEvent &e )
{
	if( e.key == OIS::KC_ESCAPE )
	{
		this->requestShutdown();
	}
}
// ============================================================================
void PlayState::mouseMoved( const OIS::MouseEvent &e )
{
	GuiManager::getSingleton().getGuiPtr()->injectMouseMove(e.state.X.abs,e.state.Y.abs, 300);
}
// ============================================================================
void PlayState::mousePressed( const OIS::MouseEvent &e, OIS::MouseButtonID id )
{
	GuiManager::getSingleton().getGuiPtr()->injectMousePress(e,id);
}
// ============================================================================
void PlayState::mouseReleased( const OIS::MouseEvent &e, OIS::MouseButtonID id )
{
	GuiManager::getSingleton().getGuiPtr()->injectMouseRelease(e,id);
}
// ============================================================================
PlayState* PlayState::getSingletonPtr( void )
{
	if( !mPlayState )
	{
		mPlayState = new PlayState();
	}

	return mPlayState;
}
// ============================================================================
